
from kensu.utils.kensu_provider import KensuProvider
from kensu.utils.dsl import mapping_strategies


def wrap_pandas_gbq_write(method):
    def wrapper(*args, **kwargs):
        kensu = KensuProvider().instance()
        df_result = method(*args, **kwargs)
        df = args[0]  # see get_dummies definition (first arg is `data`)

        orig_ds = kensu.extractors.extract_data_source(df, kensu.default_physical_location_ref,
                                                     logical_naming=kensu.logical_naming)._report()
        orig_sc = kensu.extractors.extract_schema(orig_ds, df)._report()

        result_ds = kensu.extractors.extract_data_source(df_result, kensu.default_physical_location_ref,
                                                       logical_naming=kensu.logical_naming)._report()
        result_sc = kensu.extractors.extract_schema(result_ds, df_result)._report()

        df_result_kensu = DataFrame.using(df_result)

        kensu.add_dependency((df, orig_ds, orig_sc), (df_result, result_ds, result_sc),
                           mapping_strategy=mapping_strategies.OUT_STARTS_WITH_IN)

        return df_result_kensu

    wrapper.__doc__ = method.__doc__
    return wrapper


import sklearn.metrics as mk
from kensu.numpy import ndarray
from kensu.pandas import DataFrame,Series
from kensu.utils.wrappers import remove_ksu_wrappers


def wrap_classification(method):
    def wrapper(*args, **kwargs):
        new_args=remove_ksu_wrappers(args)
        result = method(*new_args, **kwargs)
        return result

    wrapper.__doc__ = method.__doc__
    return wrapper


classification_report = wrap_classification(mk.classification_report)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
#
# Copyright 2021 Kensu Inc
#
import logging

from kensu.google.cloud.bigquery.job.bq_helpers import BqKensuHelpers
from kensu.utils.dsl.extractors.external_lineage_dtos import KensuDatasourceAndSchema, GenericComputedInMemDs, \
    ExtDependencyEntry
from kensu.utils.kensu import Kensu
import google.cloud.bigquery as bq
import sqlparse


class BqOfflineParser:

    # FIXME: or should we better simply fetch schema ALL visible tables and databases !!!!???
    @staticmethod
    def get_referenced_tables_metadata(
            kensu: Kensu,
            client: bq.Client,
            query: str):
        table_infos = BqOfflineParser.get_table_info_from_sql(client, query)
        # for table, ds, sc in table_infos:
        #     # FIXME: this possibly don't fit here well...
        #     kensu.real_schema_df[sc.to_guid()] = table

        table_id_to_bqtable = {}
        metadata = {"tables": []}
        for table, ds, sc in table_infos:
            table_id = "`" + table.full_table_id.replace(":", ".") + "`"  # FIXME: replace this in DS extractor too!
            table_md = {
                "id": table_id,
                "schema": {
                    "fields": [{"name": f.name, "type": f.field_type} for f in sc.pk.fields]
                }
            }
            table_id_to_bqtable[table_id] = table
            metadata["tables"].append(table_md)
        return metadata,  table_id_to_bqtable, table_infos

    @staticmethod
    def get_table_info_for_id(client: bq.Client, id: sqlparse.sql.Identifier):
        try:
            name = (id.get_real_name()).strip('`')
            table = client.get_table(name)
            ds, sc = BqKensuHelpers.table_to_kensu(table)  # FIXME?
            return table, ds, sc
        except Exception as e:
            logging.debug("get_table_info_for_id failed for table={}, maybe not BQ table: {}".format(id, str(e)))
            # FIXME this is because the current find_sql_identifiers also returns the column names...
            #  (see aboveREF_GET_TABLE)
            #  Therefore get_table of a column name should fail
            return None

    @staticmethod
    def get_table_info_from_sql(client: bq.Client, query: str):
        sq = sqlparse.parse(query)
        ids = BqOfflineParser.find_sql_identifiers(sq[0].tokens)  # FIXME we only take the first element
        table_infos = list(
            filter(lambda x: x is not None, [BqOfflineParser.get_table_info_for_id(client, id) for id in ids]))
        return table_infos

    @staticmethod
    def find_sql_identifiers(tokens):
        for t in tokens:
            if isinstance(t, sqlparse.sql.Identifier):
                if t.is_group and len(t.tokens) > 0:
                    # String values like "World" in `N == "World"` are also Identifier
                    # but their first child is of ttype `Token.Literal.String.Symbol`
                    # although table seems to have a first child of ttype `Token.Name`
                    if str(t.tokens[0].ttype) == "Token.Name":
                        # FIXME .. this is also returning the column names... (REF_GET_TABLE)
                        yield t
            elif t.is_group:
                yield from BqOfflineParser.find_sql_identifiers(t)

    @staticmethod
    def fallback_lineage(kensu, table_infos, dest):
        global_lineage = []
        for table, ds, sc in table_infos:
            ds_path = ds.pk.location
            schema_fields = [(f.name, f.field_type) for f in sc.pk.fields]
            input = KensuDatasourceAndSchema.for_path_with_opt_schema(
                kensu,
                ds_path=ds_path,
                ds_name=ds.name,
                format='BigQuery table',
                categories=None,
                maybe_schema=schema_fields,
                f_get_stats=None  # FIXME
            )
            lin_entry = ExtDependencyEntry(
                input_ds=input,
                lineage=dict([(v.name, v.name) for v in sc.pk.fields])  # FIXME: check if output field exists
            )
            global_lineage.append(lin_entry)
        return GenericComputedInMemDs(lineage=global_lineage)


import logging
import datetime

from kensu.google.cloud.bigquery import Client
from kensu.utils.kensu_provider import KensuProvider


def compute_bigquery_stats(table_ref, table, client, stats_aggs, input_filters=None):
    r = {}
    kensu = KensuProvider().instance()
    client: Client = client or kensu.data_collectors['BigQuery']
    if stats_aggs is None:
        logging.debug('Got empty statistic listing from remote service, proceeding with fallback statistic list')
        stats_aggs = generate_fallback_stats_queries(table)

    selector = ",".join([sql_aggregation + " " + col + "_" + stat_name
                         for col, stats_for_col in stats_aggs.items()
                         for stat_name, sql_aggregation in stats_for_col.items()])
    filters = ''
    if input_filters is not None and len(input_filters) > 0:
        filters = f"WHERE {' AND '.join(input_filters)}"
    stats_query = f"select {selector}, sum(1) as nrows from `{str(table_ref)}` {filters}"
    logging.debug(f"stats query for table {table_ref}: {stats_query}")
    for row in client.query(stats_query).result():
        # total num rows (independent of column)
        r['nrows'] = row['nrows']
        # extract column specific stats
        for col, stat_names in stats_aggs.items():
            for stat_name in stat_names.keys():
                v = row[col + "_" + stat_name]
                if v.__class__ in [datetime.date, datetime.datetime, datetime.time]:
                    v = int(v.strftime("%s") + "000")
                r[col + "." + stat_name] = v
        break  # there should be only one row here
    return r


def generate_fallback_stats_queries(table):
    stats_aggs = {}
    for f in table.schema:
        # f.field_type is
        # ["STRING", "BYTES",
        # "INTEGER", "INT64",
        # "FLOAT", "FLOAT64",
        # "BOOLEAN", "BOOL",
        # "TIMESTAMP", "DATE", "TIME", "DATETIME", "GEOGRAPHY", "NUMERIC", "BIGNUMERIC",
        # "RECORD", "STRUCT",]
        # FIXME: decimal
        if f.field_type in ["INTEGER", "INT", "FLOAT", "FLOAT64", "NUMERIC", "BIGNUMERIC"]:
            stats_aggs[f.name] = {"min": f"min({f.name})",
                                  "max": f"max({f.name})",
                                  "mean": f"avg({f.name})",
                                  "nullrows": f"sum(case {f.name} when null then 1 else 0 end)"}
        elif f.field_type in ["TIMESTAMP", "DATE", "TIME", "DATETIME"]:
            stats_aggs[f.name] = {"min": f"min({f.name})",
                                  "max": f"max({f.name})",
                                  "nullrows": f"sum(case {f.name} when null then 1 else 0 end)"}
        elif f.field_type in ["BOOLEAN", "BOOL"]:
            stats_aggs[f.name] = {"true": f"sum(case {f.name} when true then 1 else 0 end)",
                                  "nullrows": f"sum(case {f.name} when null then 1 else 0 end)"}
    return stats_aggs



import logging
import re
from hashlib import sha1

# fixme: circular import, so need to inline in each fn?
# from kensu.utils.kensu_provider import KensuProvider

from kensu.client import FieldDef, SchemaPK, Schema, DataSource


def to_snake_case(name):
    s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()


def singleton(cls, *args, **kw):
    instances = {}

    def _singleton():
        if cls not in instances:
            instances[cls] = cls(*args, **kw)
        return instances[cls]

    return _singleton


def to_hash_key(o):
    return sha1(str(o).encode()).hexdigest()


def eventually_report_in_mem(o):
    from kensu.utils.kensu_provider import KensuProvider
    kensu = KensuProvider().instance()
    if kensu.report_in_mem:
        o._report()
    return o


def get_absolute_path(path):
    import os
    for prefix in ["abfss", "/abfss", "dbfs", "/dbfs"]:
        if path.startswith(prefix):
            path = path.replace(prefix, '')
            prefix = prefix.replace('/', '')
            return prefix + ':' + path
    return 'file:' + str(os.path.abspath(path))


def maybe_report(o, report):
    from kensu.utils.kensu_provider import KensuProvider
    dam = KensuProvider().instance()
    if report:
        o._report()
    return o


def extract_ksu_ds_schema(kensu, orig_variable, report=False, register_orig_data=False):
    ds = kensu.extractors.extract_data_source(orig_variable, kensu.default_physical_location_ref, logical_naming=kensu.logical_naming)
    schema = kensu.extractors.extract_schema(ds, orig_variable)
    maybe_report(ds, report=report)
    maybe_report(schema, report=report)
    if register_orig_data:
        kensu.real_schema_df[schema.to_guid()] = orig_variable
    return ds, schema


def report_all2all_lineage(in_obj, out_obj, op_type, in_inmem=True, out_inmem=True):
    from kensu.utils.kensu_provider import KensuProvider
    kensu = KensuProvider().instance()
    in_ds, in_schema = extract_ksu_ds_schema(kensu,
                                             in_obj,
                                             report=kensu.report_in_mem or not in_inmem,
                                             register_orig_data=not in_inmem)
    out_ds, out_schema = extract_ksu_ds_schema(kensu,
                                               out_obj,
                                               report=kensu.report_in_mem or not out_inmem,
                                               register_orig_data=not out_inmem)
    logging.debug("in_schema=" + str(in_schema) + "\nout_schema=" + str(out_schema))
    if kensu.mapping:
        for col in out_obj:
            if col in [v.name for v in in_schema.pk.fields]:
                kensu.add_dependencies_mapping(guid=out_schema.to_guid(),
                                               col=str(col),
                                               from_guid=in_schema.to_guid(),
                                               from_col=str(col),
                                               type=op_type)
def flatten(d, parent_key='', sep='.'):
    items = []
    if isinstance(d,list):
        for element in d:
            items.extend(flatten(element, parent_key='[]', sep=sep).items())
    else:
        for k, v in d.items():
            new_key = parent_key + sep + k if parent_key else k
            if isinstance(v, dict):
                items.extend(flatten(v, new_key, sep=sep).items())
            elif isinstance(v, list):
                if isinstance(v[0], dict):
                    new_key = new_key + '[]'
                    for i in v:
                        items.extend(flatten(i, new_key, sep=sep).items())
            else:
                items.append((new_key, type(v).__name__))
    return dict(items)

def logical_naming_batch(string):
    from itertools import groupby, chain
    grouped = groupby(string, str.isdigit)
    return ''.join(chain.from_iterable("<number>" if k else g for k,g in grouped))


def to_datasource(ds_pk, format, location, logical_naming, name):
        if logical_naming == 'File':
            logical_category = location.split('/')[-1]
            ds = DataSource(name=name, format=format, categories=['logical::' + logical_category], pk=ds_pk)
        else:
            ds = DataSource(name=name, format=format, categories=[], pk=ds_pk)
        return ds


def extract_short_json_schema(result, result_ds):
    fields_set = set()
    if isinstance(result, list):
        for element in result:
            if isinstance(element,dict):
                for e in element.keys():
                    fields_set.add(('[].' + str(e), type((element[e])).__name__))
            else:
                fields_set.add(('value', 'unknown'))
    elif isinstance(result, dict):
        for e in result.keys():
            fields_set.add((e, type(result[e]).__name__))
    else:
        fields_set.add(('value', 'unknown'))
    fields = [FieldDef(name=k[0], field_type=k[1], nullable=True) for k in fields_set]

    sc_pk = SchemaPK(result_ds.to_ref(),fields=fields)

    short_result_sc = Schema(name="short-schema:" + result_ds.name, pk=sc_pk)
    return short_result_sc




from kensu.pandas import DataFrame,Series
from kensu.utils.kensu_provider import KensuProvider
from gluonts.dataset.common import  ListDataset
from kensu.utils.helpers import eventually_report_in_mem
from kensu.numpy import ndarray
from kensu.client.models import Model,ModelPK,ModelRef,ModelTraining,ModelTrainingPK,ModelTrainingRef





import pysftp
import os
import posixpath
import logging


# this import is actually used by end-user
from pysftp import CnOpts


class Connection(pysftp.Connection):

    def __init__(self, host, username=None, private_key=None, password=None,
                 port=22, private_key_pass=None, ciphers=None, log=False,
                 cnopts=None, default_path=None):
        super().__init__(host,
                         username=username,
                         private_key=private_key,
                         password=password,
                         port=port,
                         private_key_pass=private_key_pass,
                         ciphers=ciphers,
                         log=log,
                         cnopts=cnopts,
                         default_path=default_path)
        self.ksu_sftp_host = host
        self.ksu_sftp_port = port

    def _ftp_server_addr(self):
        return f"ftp://{self.ksu_sftp_host}:{self.ksu_sftp_port}"

    def put(
            self,
            localpath,
            remotepath=None,
            callback=None,
            confirm=True,
            preserve_mtime=False
    ):
        result = super(Connection, self).put(
            localpath,
            remotepath=remotepath,
            callback=callback,
            confirm=confirm,
            preserve_mtime=preserve_mtime
        )
        if not remotepath:
            remotepath = os.path.split(localpath)[1]
        resolved_remote_path = posixpath.join(self.pwd, remotepath)
        qualified_remote_path = self._ftp_server_addr() + resolved_remote_path
        absolute_localpath = os.path.abspath(str(localpath))

        logging.info(f"sftp.put(localpath: {absolute_localpath}, remote_qualified_path:{qualified_remote_path})")
        # schema is not known for a generic ftp file copy !!! but we try to guess it
        maybe_schema = [("unknown, unknown"), ]
        from kensu.utils.helpers import extract_ksu_ds_schema
        from kensu.utils.helpers import get_absolute_path as get_kensu_abs_path
        kensu_normalized_absolute_path = get_kensu_abs_path(absolute_localpath)
        try:
            import kensu.pandas as pd
            maybe_pandas_df = pd.read_csv(absolute_localpath, sep=";")
            from kensu.utils.kensu_provider import KensuProvider
            ksu = KensuProvider().instance()
            ds, schema = extract_ksu_ds_schema(ksu, maybe_pandas_df, report=False, register_orig_data=False)
            maybe_schema = [(f.name, f.field_type) for f in schema.pk.fields]
        except:
            print("Kensu failed to infer schema for data of pysftp copy operation...")
            import traceback
            traceback.print_exc()
            traceback.print_tb(tb=None)
        print('Extracted schema:', maybe_schema)
        from kensu.utils.dsl.extractors.external_lineage_dtos import GenericComputedInMemDs
        GenericComputedInMemDs.report_copy_with_opt_schema(
            src=kensu_normalized_absolute_path,
            dest=qualified_remote_path,
            operation_type="sftp.put()",
            maybe_schema=maybe_schema
        )

    put.__doc__ = pysftp.Connection.put.__doc__


Connection.__doc__ = pysftp.Connection.__doc__



from kensu.psycopg2.parser import parse_and_report
from psycopg2.extras import *
import psycopg2
from psycopg2.extras import _split_sql, _paginate

# FIXME: now it's copy paste now, but we kinda need that...
def execute_values(cur, sql, argslist, template=None, page_size=100, fetch=False):
    # kensu modifications start
    argslist = list(argslist)  # this might be iterator, and it'd break if accessed multiple times
    # call original function for querying db and returning result
    result = psycopg2.extras.execute_values(cur, sql, argslist, template=template, page_size=page_size,
                                            fetch=fetch)

    # below is mostly original code from psycopg2.extras.execute_values
    # - extract the first query (of possible multiple ones), and feed it to postgres SQL parser
    from psycopg2.sql import Composable
    if isinstance(sql, Composable):
        sql = sql.as_string(cur)

    # we can't just use sql % vals because vals is bytes: if sql is bytes
    # there will be some decoding error because of stupid codec used, and Py3
    # doesn't implement % on bytes.
    if not isinstance(sql, bytes):
        from psycopg2 import extensions as _ext
        sql = sql.encode(_ext.encodings[cur.connection.encoding])

    pre, post = _split_sql(sql)

    first_sql_query_bytes = None
    for page in _paginate(argslist, page_size=page_size):
        if template is None:
            template = b'(' + b','.join([b'%s'] * len(page[0])) + b')'
        parts = pre[:]
        for args in page:
            parts.append(cur.mogrify(template, args))
            parts.append(b',')
        parts[-1:] = post
        first_sql_query_bytes = b''.join(parts)
        break

    if first_sql_query_bytes is not None:
        parse_and_report(cur, first_sql_query_bytes, argslist)

    return result

execute_values.__doc__ = psycopg2.extras.execute_values.__doc__



from kensu.client import SchemaLineageDependencyDef

class EndedBuilderElement(object):
    def __init__(self, builder):
        self.builder = builder
        self.input = self.builder.input
        self.input_schema = self.builder.input_schema
        self.output = self.builder.output
        self.output_schema = self.builder.output_schema

    def toSchemaLineageDependencyDef(self):
        return SchemaLineageDependencyDef(from_schema_ref=self.builder.input_schema.to_ref(),
                                          to_schema_ref=self.builder.output_schema.to_ref(),
                                          column_data_dependencies=self.builder.data,
                                          column_control_dependencies=self.builder.control)



def remove_ksu_wrapper(arg):
    # FIXME: fix possible import issues if dependency libs not installed...
    from kensu.pandas import DataFrame
    from kensu.pandas import Series
    from kensu.numpy import ndarray

    if isinstance(arg, DataFrame):
        return arg.get_df()
    elif isinstance(arg, Series):
        return arg.get_s()
    elif isinstance(arg, ndarray):
        return arg.get_nd()
    else:
        return arg


def remove_ksu_wrappers(args):
    new_args = []
    for arg in args:
        new_args.append(remove_ksu_wrapper(arg))
    return new_args



from gluonts.dataset.common import ListDataset
# FIXME: currently if any of imported libs were missing (pandas,numpy) it'd fail!
#  => so we'd need to extract an additional abstraction
from kensu.pandas.data_frame import DataFrame,Series
from kensu.numpy import ndarray

def make_dataset_reliable(dataset):
    dep_fields = []
    old_Field = []
    if isinstance(dataset, ListDataset):
        new_Field = []
        old_Field = dataset.list_data
        for element in dataset.list_data:
            new_dict = {}
            for key in element.keys():
                item = element[key]
                if isinstance(item, DataFrame):
                    new_item = item.get_df()
                    dep_fields.append(new_item)
                elif isinstance(item, Series):
                    new_item = item.get_s()
                    dep_fields.append(new_item)
                elif isinstance(item, ndarray):
                    new_item = item.get_nd()
                    dep_fields.append(item)
                else:
                    new_item = item
                new_dict[key] = new_item
            new_Field.append(new_dict)
        dataset.list_data = new_Field
        return (dataset, old_Field, dep_fields)
    return (dataset, old_Field, dep_fields)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
#
# Copyright 2021 Kensu Inc
#
import logging
import traceback

import google

from kensu.google.cloud.bigquery.job.offline_parser import BqOfflineParser
from kensu.google.cloud.bigquery.job.remote_parser import BqRemoteParser
from kensu.pandas import DataFrame
from kensu.utils.helpers import report_all2all_lineage
from kensu.utils.kensu_provider import KensuProvider
import google.cloud.bigquery as bq
import google.cloud.bigquery.job as bqj


class QueryJob(bqj.QueryJob):
    # - client.query(sql) returns QueryJob which we patch here for tracking
    # - for both QueryJob.to_dataframe and QueryJob.result() we return a monkey-patched QueryJob
    #      having QueryJob.result set to our wrapper fn
    #  * we add a .ksu_dest field to QueryJob.result() which is internally used to connect lineage later on
    # - QueryJob.to_dataframe calls original .result() internally, so to be able to track it, currently
    # we do not call the original .to_dataframe(), but obtain .result() and then convert it to DataFrame
    # Limitations:
    # - currently bq.TableReference (temp table) is tested as destination of job (could be RowIterator too)

    @staticmethod
    def patch(job: bqj.QueryJob) -> bqj.QueryJob:
        return QueryJob.override_result(QueryJob.override_to_dataframe(job))

    @staticmethod
    def override_to_dataframe(job: bqj.QueryJob) -> bqj.QueryJob:
        f_orig_to_dataframe = job.to_dataframe

        def wrapper(*args, **kwargs):
            # instead of calling .to_dataframe(),  to be able to track lineage
            # we get raw .result() and later convert result to pandas
            # (otherwise original input info is lost inside to_dataframe fn)
            job_result = job.result()
            to_dataframe_res = job_result.to_dataframe(*args, **kwargs)
            final_result = DataFrame.using(to_dataframe_res)
            report_all2all_lineage(in_obj=job_result, out_obj=to_dataframe_res, in_inmem=True, out_inmem=True,
                                   op_type='to_dataframe')
            return final_result

        wrapper.__doc__ = f_orig_to_dataframe.__doc__
        setattr(job, 'to_dataframe', wrapper)
        return job

    @staticmethod
    def override_result(job: bqj.QueryJob) -> bqj.QueryJob:
        f = job.result

        def wrapper(*args, **kwargs):
            kensu = KensuProvider().instance()
            result = f(*args, **kwargs)
            client = kensu.data_collectors['BigQuery']
            dest = job.destination
            if isinstance(dest, bq.TableReference):
                dest = client.get_table(dest)
            db_metadata, table_id_to_bqtable, table_infos = BqOfflineParser.get_referenced_tables_metadata(
                kensu=kensu,
                client=client,
                query=job.query)
            try:
                bq_lineage = BqRemoteParser.parse(
                    kensu=kensu,
                    client=client,
                    query=job.query,
                    db_metadata=db_metadata,
                    table_id_to_bqtable=table_id_to_bqtable)
            except:
                logging.warning("Error in BigQuery collector, using fallback implementation")
                traceback.print_exc()
                bq_lineage = BqOfflineParser.fallback_lineage(kensu, table_infos, dest)

            QueryJob._store_bigquery_job_destination(result=result, dest=dest)

            bq_lineage.report(
                ksu=kensu,
                df_result=result,
                operation_type='BigQuery SQL result',
                report_output=kensu.report_in_mem,
                # FIXME: how to know when in mem or when bigquery://projects/psyched-freedom-306508/datasets/_b63f45da1cafbd073e5c2770447d963532ac43ec/tables/anonc79d9038a13ab2dbe40064636b0aceedc62b5d69
                register_output_orig_data=False  # FIXME? when do we need this? INSERT INTO?
            )
            # FIXME? kensu.report_with_mapping()

            return result

        wrapper.__doc__ = f.__doc__
        setattr(job, 'result', wrapper)
        return job

    @staticmethod
    def _store_bigquery_job_destination(result, dest):
        try:
            if (dest is not None) and isinstance(dest, google.cloud.bigquery.table.Table):
                result.ksu_dest = dest
        except:
            logging.warning('setting _store_bigquery_job_destination failed')
            traceback.print_exc()



# coding: utf-8

"""


    No description provided (generated by Swagger Codegen https://github.com/swagger-api/swagger-codegen)

    OpenAPI spec version: beta

    Generated by: https://github.com/swagger-api/swagger-codegen.git
"""

import setuptools
from setuptools import setup, find_packages

NAME = "kensu"


VERSION = "1.5.1.1"


# To install the library, run the following
#
# python setup.py install
#
# prerequisite: setuptools
# http://pypi.python.org/pypi/setuptools

#REQUIRES = ["urllib3 >= 1.15", "six >= 1.10", "certifi", "python-dateutil"]

setup(
    name=NAME,
    version=VERSION,
    description="",
    author_email="",
    url="",
    keywords=["Ingestion", "Kensu", "Data Intelligence Manager","Analytics Observability","Data Observability"],
#    install_requires=REQUIRES,
    packages=[
        package
        for package in setuptools.PEP420PackageFinder.find()
        if package.startswith("kensu")
    ],
    platforms="Posix; MacOS X; Windows",
    include_package_data=True,
    long_description="""\
    Python Client to Report Entities to Kensu Data Activity Manager\
    """
)



import logging
import json
import time
import re
from kensu.utils.helpers import singleton, to_snake_case
from .kensu_class_handlers import KensuClassHandlers
from .. import client


@singleton
class Injection(object):
    ENTITIES_INJECTED = False
    DO_REPORT = True
    REPORT_TO_FILE = False
    OFFLINE_FILE = None

    # function taking three args: the entity, the api client, and the reporting method
    REPORTER = None

    @staticmethod
    def do_nothing_reporter(obj, kensu_api, method):
        return obj

    @staticmethod
    def printing_reporter(obj, kensu_api, method):
        json = Injection().get_offline_entity_json(obj, kensu_api)
        print(json)
        return obj

    @staticmethod
    def logging_info_reporter(obj, kensu_api, method):
        json = Injection().get_offline_entity_json(obj, kensu_api)
        logging.info(json)
        return obj

    @staticmethod
    def post_reporter(obj, kensu_api, method):
        method(obj)
        return obj

    @staticmethod
    def file_reporter_builder(offline_file_name):
        inj = Injection()
        file = open(offline_file_name or "kensu_offline_events.txt", "a")
        import atexit
        atexit.register(file.close)

        def f(obj, kensu_api, method):
            inj.OFFLINE_FILE.write(inj.get_offline_entity_json(obj, kensu_api))
            inj.OFFLINE_FILE.flush()
            return obj

        return f


    def __init__(self):
        pass

    def set_reporter(self, reporter):
        self.REPORTER = reporter

    def set_do_report(self, do_report=True, report_to_file=False, offline_file_name=None):
        self.DO_REPORT = do_report
        self.REPORT_TO_FILE = report_to_file
        if report_to_file:
            self.OFFLINE_FILE = open(offline_file_name or "kensu_offline_events.txt", "a")
            import atexit
            atexit.register(self.close_gracefully)

    def close(self):
        if self.OFFLINE_FILE is not None:
            self.OFFLINE_FILE.close()

    def close_gracefully(self):
        try:
            self.close()
        except IOError as ex:
            print('Failed closing the offline output file', ex)

    def set_kensu_api(self, kensu_api):
        if not self.ENTITIES_INJECTED:
            self.kensu_inject_entities(kensu_api)

    # injects to_ref and _report to Kensu entities
    def kensu_inject_entities(self, kensu_api):
        if self.ENTITIES_INJECTED:
            return
        this_injection = self
        kensu_entities = ["DataSource", "ProcessLineage", "Process", "ProcessRun", "User",
                        "Schema", "CodeVersion", "CodeBase", "PhysicalLocation", "Model",
                        "ModelTraining", "ModelMetrics", "DataStats", "LineageRun",
                        "Project"]

        def kensu_to_ref(self, use_pk=False):
            if use_pk:
                return getattr(client, self.__class__.__name__ + "Ref")(by_pk=self.pk)
            else:
                return getattr(client, self.__class__.__name__ + "Ref")(by_guid=self.to_guid())

        # note: self here is not Injection anymore but the object used during the method call below
        def kensu_report(self):
            c = self.__class__.__name__
            cc = to_snake_case(c)
            method_name = "report_" + cc
            method = getattr(kensu_api, method_name)
            if this_injection.REPORTER:
                this_injection.REPORTER(self, kensu_api, method)
            elif this_injection.DO_REPORT:
                if not this_injection.REPORT_TO_FILE:
                    # call method with self
                    method(self)
                else:
                    Injection().OFFLINE_FILE.write(this_injection.get_offline_entity_json(self, kensu_api))
                    Injection().OFFLINE_FILE.flush()
            return self

        to_guid_func = lambda x: KensuClassHandlers.guid.__func__(KensuClassHandlers, x)
        for c in kensu_entities:
            setattr(getattr(client, c), 'to_guid', to_guid_func)
            setattr(getattr(client, c), 'to_ref', kensu_to_ref)
            setattr(getattr(client, c), '_report', kensu_report)

        self.ENTITIES_INJECTED = True

    @staticmethod
    def get_offline_entity_json(entity, kensu_api):
        # FIXME: check if there's a nicer way than to call kensu_api.api_client.sanitize_for_serialization
        sanitized_body = kensu_api.api_client.sanitize_for_serialization(entity)
        # p.s. serialization based on existing rules in kensu-client-scala
        now = round(1000.0 * time.time())
        offline_entity = {
            "action": "add_entity",
            "entity": re.sub(r'(.)([A-Z])', r'\1_\2', entity.__class__.__name__).upper(),
            "generatedEntityGUID": "empty",
            "schemaVersion": "0.1",
            "jsonPayload": sanitized_body,
            "context": {
                "clientId": "",
                # FIXME: make this configurable for reporting events in past
                "clientEventTimestamp": now,
                "serverReceivedTimestamp": now
            }
        }
        return json.dumps(offline_entity) + "\n"
